/* A somewhat amsart-like style */

.ltx_title_document {
    text-transform:uppercase; font-size: 150%; text-align:center; }
.ltx_title_section {
    font-family:serif; font-weight:normal; font-variant:small-caps; font-size: 120%;
    text-align:center; }
.ltx_authors { text-align:center; }
.ltx_author_before,.ltx_author_after,
.ltx_role_author .ltx_personname {
    text-transform:uppercase; }
.ltx_abstract { margin-bottom:1em; }
.ltx_title_abstract { font-family:serif; font-weight:normal; font-variant:small-caps;}
.ltx_title_keywords,
.ltx_title_classification { display:inline; font-size:100%; font-weight:normal; font-style:italic; }
.ltx_role_dedicatory { font-size:100%; font-style:italic; text-align:center; margin:1em;}

.ltx_figure .ltx_tag,
.ltx_table .ltx_tag  {
    font-family:serif; font-variant:small-caps;}

.ltx_title_subsection,
.ltx_title_subsubsection {
    font-family:serif; font-weight:bold;}
.ltx_title_subsection:after,
.ltx_title_subsubsection:after,
.ltx_title_paragraph:after  { content:"."; }

/* abstract, subsection, subsubsection and paragraph titles are runin */
.ltx_abstract .ltx_title,
.ltx_abstract .ltx_title + .ltx_p { display:inline; }
.ltx_subsection .ltx_title,
.ltx_subsection .ltx_title + .ltx_para,
.ltx_subsection .ltx_title + .ltx_para > .ltx_p { display:inline; }
.ltx_subsubsection .ltx_title,
.ltx_subsubsection .ltx_title + .ltx_para,
.ltx_subsubsection .ltx_title + .ltx_para > .ltx_p { display:inline; }
.ltx_paragraph .ltx_title,
.ltx_paragraph .ltx_title + .ltx_para,
.ltx_paragraph .ltx_title + .ltx_para > .ltx_p { display:inline; }

.ltx_title +.ltx_para > .ltx_p {text-indent:2em; }

.ltx_itemize,
.ltx_enumerate,
.ltx_description
    { margin-left:2em; }
