/* CSS for left navigation bar */

.ltx_page_navbar {
    display:block!important; position:fixed; left:0px; top:0px; width:170px;
    margin:0em; padding:1em; font: bold 75% sans-serif; 
    border: 3px double; }
.ltx_page_navbar ul {
    margin-left:-2em; }
.ltx_page_main {
    position:absolute; left:190px; top:0px; right:2px; 
    margin:0px; padding:1em 3em 1em 2em;
    width:70%; }
