/* Blue theme */
.ltx_page_navbar  { border       :5px double #485C5A; background-color:#8C9C9A; }
.ltx_page_header  { border-bottom:1px solid  #485C5A; background-color:#8C9C9A; }
.ltx_page_footer  { border-top   :1px solid  #485C5A; background-color:#8C9C9A; }

.ltx_title_document,
.ltx_title_chapter,
.ltx_title_section   { border-top : 5px double #485C5A;
                       border-left: 3px double #485C5A;
                       background-color:#9DB2B1; }
.ltx_title_subsection,
.ltx_title_subsubsection,
.ltx_title_paragraph { border-top : 1px solid #485C5A;
                       border-left: 1px solid #485C5A;
                       background-color:#BFCFCC; }

th, *[class~="thead"] { background-color:#8C9C9A; }

.ltx_page_content { background-color:#D6E4E1; }