/*======================================================================
 Core CSS for LaTeXML documents converted to (X)HTML */
/* Generic Page layout */
.ltx_page_header,
.ltx_page_footer { font-size:0.8em; }
.ltx_page_header *[rel~="prev"],
.ltx_page_footer *[rel~="prev"] { float:left; }
.ltx_page_header *[rel~="up"],
.ltx_page_footer *[rel~="up"]   { display:block; text-align:center; }
.ltx_page_header *[rel~="next"],
.ltx_page_footer *[rel~="next"] {  float:right; }
/* What was I trying for here; need more selective rule!
.ltx_page_header .ltx_ref,
.ltx_page_footer .ltx_ref {
    margin:0 1em; }
*/
.ltx_page_header li {
    padding:0.1em 0.2em 0.1em 1em;}

/* Main content */
.ltx_page_content { clear:both; }
.ltx_page_header  { border-bottom:1px solid; margin-bottom:5px; }
.ltx_page_footer  { clear:both; border-top:1px solid; margin-top:5px;  }

.ltx_page_header:after,
.ltx_page_footer:after,
.ltx_page_content:after { content:"."; display:block; height:0; clear:both; visibility:hidden; }

.ltx_page_footer:before
 { content:"."; display:block; height:0; clear:both; visibility:hidden; }

.ltx_page_logo     { font-size:80%; margin-top: 5px; clear:both; float:right; }
.ltx_page_logo a   { font-variant: small-caps; }
.ltx_page_logo img { vertical-align:-3px; }

/* if shown */
.ltx_page_navbar li { white-space:nowrap; display:block; overflow:hidden; }
/* If ref got turned into span, it's "this section"*/
.ltx_page_navbar li span.ltx_ref { white-space:normal; overflow:visible; }

/* Ought to be easily removable/overridable? */
.ltx_pagination.ltx_role_newpage { height:2em; }
/*======================================================================
  Document Structure; Titles & Frontmatter */

/* undo bold here to remove the browser's native h# styling,
   at let all other styles override it (with more specific rules)*/
.ltx_title { font-size:100%; font-weight:normal; }

/* Hack to simulate run-in! put class="ltx_runin" on a title or tag
   for it to run-into the following text. */
.ltx_runin { display:inline; }
.ltx_runin:after { content:" "; }
.ltx_runin + .ltx_para,
.ltx_runin + .ltx_para p,
.ltx_runin + p { display:inline; }

.ltx_outdent { margin-left: -2em; }

/* .ltx_chapter_title, etc should be in ltx-article.css etc.
 */
.ltx_page_main { margin:0px; padding:1em 3em 1em 2em; }
.ltx_tocentry  { list-style-type:none; }

/* support for common author block layouts.*/
/* add class ltx_authors_1line to get authors in single line
   with pop-up affiliation, etc. */
.ltx_authors_1line .ltx_creator,
.ltx_authors_1line .ltx_author_before,
.ltx_authors_1line .ltx_author_after { display:inline;}
.ltx_authors_1line .ltx_author_notes { display:inline-block; }
.ltx_authors_1line .ltx_author_notes:before { content:"*"; color:blue;}
.ltx_authors_1line .ltx_author_notes span { display:none; }
.ltx_authors_1line .ltx_author_notes:hover span {
       display:block; position:absolute; z-index:10;
       background:#E0E0E0; border:3px outset gray;
       text-align:left; }  

/* add class=ltx_authors_multiline to get authors & affliations on separate lines*/
.ltx_authors_multiline .ltx_creator,
.ltx_authors_multiline .ltx_author_before,
.ltx_authors_multiline .ltx_author_after,
.ltx_authors_multiline .ltx_author_notes,
.ltx_authors_multiline .ltx_author_notes .ltx_contact {
        display:block; }


/*======================================================================
  Para level */
td.ltx_subfigure,
td.ltx_subtable,
td.ltx_subfloat { width:50%; }
/* theorems, figure, tables, floats captions.. */
/*======================================================================
 Blocks, Lists, Floats */
.ltx_p,
.ltx_quote,
.ltx_block,
.ltx_listingblock,
.ltx_para
  { display: block; }

.ltx_align_left     {text-align:left; }
.ltx_align_right    {text-align:right; }
.ltx_align_center   {text-align:center; }
.ltx_align_justify  {text-align:justify; }
.ltx_align_top      {vertical-align:top; }
.ltx_align_bottom   {vertical-align:bottom; }
.ltx_align_middle   {vertical-align:middle; }
.ltx_align_baseline {vertical-align:baseline; }

.ltx_align_floatleft  { float:left; }
.ltx_align_floatright { float:right; }

td.ltx_align_left,   th.ltx_align_left,
td.ltx_align_right,  th.ltx_align_right,
td.ltx_align_center, th.ltx_align_center { white-space:nowrap; }

.ltx_inline-block { display:inline-block; }
div.ltx_equation { display:block; width:95%; text-align:center; }
/*.ltx_equation span.ltx_refnum.ltx_left { position:absolute; left:2em; }
.ltx_equation span.ltx_refnum.ltx_right { position:absolute; right:2em; }
*/
.ltx_tag_equation.ltx_align_left { position:absolute; left:3em; }
.ltx_tag_equation.ltx_align_right { position:absolute; right:3em; }

.ltx_equation td { width:auto; }
table.ltx_equation,
table.ltx_equationgroup { width:100%; }
table.ltx_eqn_align tr.ltx_equation td.ltx_align_left + td.ltx_align_right { padding-left:3em; }
table.ltx_eqn_eqnarray tr.ltx_eqn_lefteqn + tr td.ltx_align_right { min-width:2em; }
td.ltx_eqn_eqno { max-width:0em; overflow:visible; }
td.ltx_eqn_eqno.ltx_align_right .ltx_tag { float:right; }

/* Hide this from IE */
tr > td.ltx_eqn_center_padleft,
tr > td.ltx_eqn_center_padright { width:50%; }
tr > td.ltx_eqn_left_padleft,
tr > td.ltx_eqn_right_padright { min-width:2em; }
tr > td.ltx_eqn_left_padright,
tr > td.ltx_eqn_right_padleft  { width:100%; }

.ltx_itemize,
.ltx_enumerate,
.ltx_description
   { display:block; }
.ltx_itemize .ltx_item,
.ltx_enumerate .ltx_item
   { display: list-item; }

/* Position the tag to look like a normal item bullet. */
li.ltx_item > .ltx_tag {
   display:inline-block; margin-left:-1.5em; min-width:1.5em;
   text-align:right; }
.ltx_item .ltx_tag + .ltx_para,
.ltx_item .ltx_tag + .ltx_para .ltx_p  { display:inline; }

/* NOTE: Need to try harder to get runin appearance? */
dl.ltx_description dt { margin-right:0.5em; float:left;
                    font-weight:bold; font-size:95%; }
dl.ltx_description dd { margin-left:5em; }
dl.ltx_description dl.ltx_description dd { margin-left:3em; }

.ltx_theorem  {margin:1em 0em 1em 0em; }
.ltx_title_theorem { font-size:100%; }
.ltx_bibliography dt { margin-right:0.5em; float:left; }
.ltx_bibliography dd { margin-left:3em; }
/*.ltx_biblist { list-style-type:none; }*/
.ltx_bibitem { list-style-type:none; }
.ltx_bibtag { font-weight:bold; margin-left:-2em; width:3em; }
/*.bibitem-tag + div { display:inline; }*/
.ltx_bib_title { font-style:italic; }
.ltx_bib_article .bib-title { font-style:normal !important; }
.ltx_bib_journal  { font-style:italic; }
.ltx_bib_volume { font-weight:bold; }

.ltx_indexlist li { list-style-type:none;  }
.ltx_indexlist { margin-left:1em; padding-left:1em;}
.ltx_listing td.ltx_linenumber,
.ltx_listingblock td.ltx_linenumber
   { width:3em; text-align:right;}

.ltx_parbox {text-indent:0em; }

/* NOTE that it is CRITICAL to put position:relative outside & absolute inside!!
   I wish I understood why!
   Outer box establishes resulting size, neutralizes any outer positioning, etc;
   inner establishes position of stuff to be rotated */
.ltx_transformed_outer {
    position:relative; top:0pt;left:0pt;
    overflow:visible; }
.ltx_transformed_inner {
    display:block;
    position:absolute;top:0pt;left:0pt; }
.ltx_transformed_inner > .ltx_p {text-indent:0em; margin:0; padding:0; }

/* by default, p doesn't indent */
.ltx_p { text-indent:0em; white-space:normal; }
/* explicit control of indentation (on ltx_para) */
.ltx_indent > .ltx_p:first-child { text-indent:2em!important; }
.ltx_noindent > .ltx_p:first-child { text-indent:0em!important; }

/*======================================================================
  Columns */
.ltx_page_column1 {
    width:44%; float:left; } /* IE uses % of wrong container*/
.ltx_page_column2 {
    width:44%; float:right; }
.ltx_page_columns > .ltx_page_column1 {
    width:48%; float:left; }
.ltx_page_columns > .ltx_page_column2 {
    width:48%; float:right; }
.ltx_page_columns:after {
    content:"."; display:block; height:0; clear:both; visibility:hidden; }

/*======================================================================
 Borders and such */
.ltx_tabular { display:table; border-collapse:collapse; }
span.ltx_tabular { display:inline-table; border-collapse:collapse; }
.ltx_thead,
.ltx_tbody   { display:table-row-group; }
.ltx_tr      { display:table-row; }
.ltx_td      { display:table-cell; }

.ltx_framed  { border:1px solid black;}
.ltx_tabular td, .ltx_tabular th { padding:0.1em 0.5em; }
.ltx_border_t  { border-top:1px solid black; }
.ltx_border_r  { border-right:1px solid black; }
.ltx_border_b  { border-bottom:1px solid black; }
.ltx_border_l  { border-left:1px solid black; }
.ltx_border_tt { border-top:3px double black; }
.ltx_border_rr { border-right:3px double black; }
.ltx_border_bb { border-bottom:3px double black; }
.ltx_border_ll { border-left:3px double black; }

/*======================================================================
 Misc */
/* .ltx_verbatim*/
.ltx_verbatim { text-align:left; }
/*======================================================================
 Meta stuff, footnotes */
.ltx_note_content { display:none; }
/*right:5%;  */
.ltx_note_content {
     max-width: 70%; font-size:80%; left:15%;
     text-align:left;
     background:#E0E0E0; border:3px outset gray; }
.ltx_note_mark    { color:blue; }
.ltx_note_type    { font-weight: bold; }
.ltx_note { display:inline-block; text-indent:0; } /* So we establish containing block */
.ltx_note:hover .ltx_note_content 
   { display:block; position:absolute; z-index:10; }

.ltx_ERROR        { color:red; }
.ltx_rdf          { display:none; }

/*======================================================================
 SVG (pgf/tikz ?) basics */

/* Stuff appearing in svg:foreignObject */
.ltx_svg_fog foreignObject  { margin:0; padding:0; overflow:visible; }
.ltx_svg_fog foreignObject > p { margin:0; padding:0; display:block; }
/*.ltx_svg_fog foreignObject > p { margin:0; padding:0; display:block; white-space:nowrap; }*/

/*======================================================================
 Low-level Basics */
.ltx_font_serif      { font-family: serif; }
.ltx_font_sansserif  { font-family: sans-serif; }
.ltx_font_typewriter { font-family: monospace; }
.ltx_font_bold       { font-weight: bold; }
.ltx_font_medium     { font-weight: normal; }
.ltx_font_italic     { font-style: italic; }
.ltx_font_upright    { font-style: normal; }
.ltx_font_slanted    { font-style: oblique; }
.ltx_font_smallcaps  { font-variant: small-caps; }
.ltx_font_oldstyle   { font-variant: oldstyle-nums;  /* experimental css3 ? Doesn't seem to work!*/
                       -moz-font-feature-settings: "onum";
                       -ms-font-feature-settings: "onum";
                       -webkit-font-feature-settings: "onum";
                       font-variant-numeric: oldstyle-nums; }
.ltx_font_mathcaligraphic { font-family: "Lucida Calligraphy", "Zapf Chancery","URW Chancery L"; }
/*

.ltx_font_mathscript { ? }
*/
cite                 { font-style: normal; }

.ltx_red        { color:red; }
/*.ltx_centering  { text-align:center; margin:auto; }*/
/*.ltx_inline-block.ltx_centering,*/
/* Hmm.... is this right in general? */
.ltx_centering  { display:block; margin:auto; text-align:center; }
table.ltx_centering { display:table; } /*!!! */

/* .ltx_phantom handled in xslt */

